//
//    Copyright (C) 2011 Sascha Ittner <sascha.ittner@modusoft.de>
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
//

#include "lcec.h"
#include "lcec_el3314.h"

typedef struct {
  hal_bit_t *error;
  hal_bit_t *overrange;
  hal_bit_t *underrange;
  hal_s32_t *raw_val;
  hal_float_t *val;
  unsigned int ovr_pdo_os;
  unsigned int ovr_pdo_bp;
  unsigned int udr_pdo_os;
  unsigned int udr_pdo_bp;
  unsigned int error_pdo_os;
  unsigned int error_pdo_bp;
  unsigned int sync_err_pdo_os;
  unsigned int sync_err_pdo_bp;
  unsigned int val_pdo_os;
} lcec_el3314_chan_t;

static const lcec_pindesc_t slave_pins[] = {
  { HAL_BIT, HAL_OUT, offsetof(lcec_el3314_chan_t ,error), "%s.%s.%s.ain-%d-error" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el3314_chan_t ,overrange), "%s.%s.%s.ain-%d-overrange" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el3314_chan_t ,underrange), "%s.%s.%s.ain-%d-underrange" },
  { HAL_S32, HAL_OUT, offsetof(lcec_el3314_chan_t ,raw_val), "%s.%s.%s.ain-%d-raw" },
  { HAL_FLOAT, HAL_OUT, offsetof(lcec_el3314_chan_t ,val), "%s.%s.%s.ain-%d-val" },
  { HAL_TYPE_UNSPECIFIED, HAL_DIR_UNSPECIFIED, -1, NULL }
};

typedef struct {
  lcec_el3314_chan_t chans[LCEC_EL3314_CHANS];
} lcec_el3314_data_t;


static ec_pdo_entry_info_t lcec_el3314_pdo_entries[] = {
    {0x6000, 0x01, 1}, /* Underrange */
    {0x6000, 0x02, 1}, /* Overrange */
    {0x6000, 0x03, 2}, /* Limit 1 */
    {0x6000, 0x05, 2}, /* Limit 2 */
    {0x6000, 0x07, 1}, /* Error */
    {0x0000, 0x00, 7}, /* Gap */
    {0x6000, 0x0f, 1}, /* TxPDO State */
    {0x1800, 0x09, 1},
    {0x6000, 0x11, 16}, /* Value */
    {0x6010, 0x01, 1}, /* Underrange */
    {0x6010, 0x02, 1}, /* Overrange */
    {0x6010, 0x03, 2}, /* Limit 1 */
    {0x6010, 0x05, 2}, /* Limit 2 */
    {0x6010, 0x07, 1}, /* Error */
    {0x0000, 0x00, 7}, /* Gap */
    {0x6010, 0x0f, 1}, /* TxPDO State */
    {0x1801, 0x09, 1},
    {0x6010, 0x11, 16}, /* Value */
    {0x6020, 0x01, 1}, /* Underrange */
    {0x6020, 0x02, 1}, /* Overrange */
    {0x6020, 0x03, 2}, /* Limit 1 */
    {0x6020, 0x05, 2}, /* Limit 2 */
    {0x6020, 0x07, 1}, /* Error */
    {0x0000, 0x00, 7}, /* Gap */
    {0x6020, 0x0f, 1}, /* TxPDO State */
    {0x1802, 0x09, 1},
    {0x6020, 0x11, 16}, /* Value */
    {0x6030, 0x01, 1}, /* Underrange */
    {0x6030, 0x02, 1}, /* Overrange */
    {0x6030, 0x03, 2}, /* Limit 1 */
    {0x6030, 0x05, 2}, /* Limit 2 */
    {0x6030, 0x07, 1}, /* Error */
    {0x0000, 0x00, 7}, /* Gap */
    {0x6030, 0x0f, 1}, /* TxPDO State */
    {0x1803, 0x09, 1},
    {0x6030, 0x11, 16}, /* Value */
};

static ec_pdo_info_t lcec_el3314_pdos[] = {
    {0x1a00, 9, lcec_el3314_pdo_entries + 0}, /* TC TxPDO-MapCh.1 */
    {0x1a01, 9, lcec_el3314_pdo_entries + 9}, /* TC TxPDO-MapCh.2 */
    {0x1a02, 9, lcec_el3314_pdo_entries + 18}, /* TC TxPDO-MapCh.3 */
    {0x1a03, 9, lcec_el3314_pdo_entries + 27}, /* TC TxPDO-MapCh.4 */
};

static ec_sync_info_t lcec_el3314_syncs[] = {
    {0, EC_DIR_OUTPUT, 0, NULL, EC_WD_DISABLE},
    {1, EC_DIR_INPUT, 0, NULL, EC_WD_DISABLE},
    {2, EC_DIR_OUTPUT, 0, NULL, EC_WD_DISABLE},
    {3, EC_DIR_INPUT, 4, lcec_el3314_pdos + 0, EC_WD_DISABLE},
    {0xff}
};

void lcec_el3314_read(struct lcec_slave *slave, long period);

int lcec_el3314_init(int comp_id, struct lcec_slave *slave, ec_pdo_entry_reg_t *pdo_entry_regs) {
  lcec_master_t *master = slave->master;
  lcec_el3314_data_t *hal_data;
  lcec_el3314_chan_t *chan;
  int i;
  int err;

  // initialize callbacks
  slave->proc_read = lcec_el3314_read;

  // alloc hal memory
  if ((hal_data = hal_malloc(sizeof(lcec_el3314_data_t))) == NULL) {
    rtapi_print_msg(RTAPI_MSG_ERR, LCEC_MSG_PFX "hal_malloc() for slave %s.%s failed\n", master->name, slave->name);
    return -EIO;
  }
  memset(hal_data, 0, sizeof(lcec_el3314_data_t));
  slave->hal_data = hal_data;

  // initializer sync info
  slave->sync_info = lcec_el3314_syncs;  
  
  // initialize pins
  for (i=0; i<LCEC_EL3314_CHANS; i++) {
    chan = &hal_data->chans[i];
		
    // initialize POD entries
    LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6000 + (i << 4), 0x01, &chan->ovr_pdo_os, &chan->ovr_pdo_bp);
    LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6000 + (i << 4), 0x02, &chan->udr_pdo_os, &chan->udr_pdo_bp );
    LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6000 + (i << 4), 0x07, &chan->error_pdo_os, &chan->error_pdo_bp);
    LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6000 + (i << 4), 0x11, &chan->val_pdo_os, NULL);

    // export pins
    if ((err = lcec_pin_newf_list(chan, slave_pins, LCEC_MODULE_NAME, master->name, slave->name, i)) != 0) {
      return err;
    }

    // initialize pins
    *(chan->error) = 0;
    *(chan->overrange) = 0;
    *(chan->underrange) = 0;
    *(chan->raw_val) = 0;
    *(chan->val) = 0.0;
  }

  return 0;
}

void lcec_el3314_read(struct lcec_slave *slave, long period) {
  lcec_master_t *master = slave->master;
  lcec_el3314_data_t *hal_data = (lcec_el3314_data_t *) slave->hal_data;
  uint8_t *pd = master->process_data;
  int i;
  lcec_el3314_chan_t *chan;
  uint8_t state;
  int16_t value;

  // wait for slave to be operational
  if (!slave->state.operational) {
    return;
  }

  chan = &hal_data->chans[0];
  *(chan->underrange) =  EC_READ_S16(&pd[chan->udr_pdo_os]);

  chan = &hal_data->chans[1];
  *(chan->underrange) =  EC_READ_BIT(&pd[chan->udr_pdo_os], chan->udr_pdo_bp);

  // check inputs
  for (i=0; i<LCEC_EL3314_CHANS; i++) {
    chan = &hal_data->chans[i];

    // update state
    *(chan->error) = EC_READ_BIT(&pd[chan->error_pdo_os], chan->error_pdo_bp);
    *(chan->overrange) = EC_READ_BIT(&pd[chan->ovr_pdo_os], chan->ovr_pdo_bp);
    *(chan->underrange) =  EC_READ_BIT(&pd[chan->udr_pdo_os], chan->udr_pdo_bp);

    // update value
    value = EC_READ_S16(&pd[chan->val_pdo_os]);
    *(chan->raw_val) = value;
    *(chan->val) = (double)value*0.1;
  }
}


